from __future__ import print_function, division
import ee
from hydrafloods import decorators


def starfm(
    coarseCollection, fineCollection=None, targetDate="1970-01-01", windowSize=33, A=0.5
):
    @decorators.carry_metadata
    def apply_starfm(img):
        t = ee.Date(img.get("system:time_start"))

        Tijk = img.select("time").subtract(base.select("time_first"))

        Sijk = img.subtract(slv).abs().convolve(Dijk).rename(bandList)

        Cijk = Sijk.multiply(Tijk).convolve(Dijk)
        Wijk = one.divide(Cijk).divide(one.divide(Cijk).convolve(Dijk))

        nBands = ee.Number(base.bandNames().length())
        expected = ee.Number(len(bandList))

        outImg = ee.Algorithms.If(
            nBands.neq(expected),
            None,
            base.divide(Wijk.convolve(Dijk))
            .add(Sijk)
            .set("system:time_start", t)
            .rename(bandList),
        )

        return ee.Image(outImg)

    target = ee.Date(targetDate)

    if type(fineCollection) is not ee.imagecollection.ImageCollection:
        try:
            fineCollection = getattr(fineCollection, "collection")
        except Exception as e:
            raise TypeError(
                "keyword fineCollection needs to be either of type ee.ImageCollection "
                "or hf.hfCollection"
            )

    bandList = ee.Image(coarseCollection.first()).bandNames().getInfo()

    one = ee.Image.constant(1)
    centerPos = ee.Number((windowSize - 1) / 2)
    pos = ee.Array(ee.List.sequence(0, windowSize - 1))
    xPos = pos.repeat(1, windowSize)
    yPos = pos.repeat(1, windowSize).transpose()

    dijk = (
        ee.Array(centerPos)
        .subtract(xPos)
        .pow(2)
        .add(ee.Array(centerPos).subtract(yPos).pow(2))
        .sqrt()
    )

    weightMax = dijk.get([0, 0]).add(1)
    # flip weights to have heighest in middle
    # normalize values for convolutions
    dijk_norm = dijk.subtract(weightMax).abs().divide(weightMax.subtract(1))

    dW = ee.Array(1).add(dijk.divide(ee.Array(A)))

    square = ee.Kernel.square(windowSize, "pixels")
    Dijk = ee.Kernel.fixed(
        windowSize, windowSize, dW.toList(), centerPos, centerPos, True
    )

    base = (
        fineCollection.filterDate(target.advance(-2, "month"), target)
        .sort("system:time_start", False)
        .reduce(ee.Reducer.firstNonNull())
    )

    slv = (
        coarseCollection.filterDate(target.advance(-2, "month"), target)
        .sort("system:time_start", False)
        .reduce(ee.Reducer.firstNonNull())
    )

    result = coarseCollection.filterDate(
        target.advance(-5, "day"), target.advance(5, "day")
    ).map(apply_starfm, True)

    return result


def bathtub(wfrac, hand, permanent=None):
    """
    Function to fit hi-resolution HAND model to water fraction estimate
    args:
        wFrac (ee.Image): water fraction image, values must be 0-1
        hand (ee.Image): height above nearest drainage (HAND) image, units in meters
        permanent (ee.Image): permanent water image to seed HAND filling
    """

    def fillGrids(d):
        d = ee.Number(d)
        dMap = (
            hand.lte(d)
            .Or(permWater)
            .reduceResolution(
                reducer=ee.Reducer.mean(), bestEffort=True, maxPixels=1024
            )
            .reproject(crs=proj)
        )

        diff = wfrac.subtract(dMap).abs()
        return diff

    def minimizeDepth(d):
        d = ee.Number(d)
        dMap = hand.lte(d).Or(permWater)
        rd = dMap.reduceResolution(
            reducer=ee.Reducer.mean(), bestEffort=True, maxPixels=1024
        ).reproject(crs=proj)

        diff = wfrac.subtract(rd).abs()

        out = nil.where(diff.eq(minDiff), dMap)
        err = nil.where(diff.eq(minDiff), diff)
        return out.rename("water").addBands(err.rename("error"))

    nil = ee.Image(0)
    if permanent:
        permWater = permanent
    else:
        permWater = ee.Image(0)
    proj = wfrac.projection()

    depths = ee.List.sequence(0, 20)

    depthMaps = ee.ImageCollection(depths.map(fillGrids))
    minDiff = depthMaps.min()

    waterMap = ee.ImageCollection(depths.map(minimizeDepth))
    final = (
        waterMap.select("water").max().addBands(waterMap.select("error").min().float())
    )

    return final



from __future__ import absolute_import
import os
import ee
import fire
import glob
import yaml
import datetime
import warnings
import pandas as pd
import geopandas as gpd

from hydrafloods import utils
from hydrafloods.processing import *

class hydrafloods(object):
    def __init__(self,config=None):
        self.configuration = config
        return

    def _parse_config(self):
        if self.configuration:
            self.filePath = os.path.dirname(os.path.abspath(__file__))
            yamlFile = self.configuration

            with open(self.configuration,'r') as stream:
                try:
                    struct = yaml.load(stream,Loader=yaml.SafeLoader)
                except yaml.YAMLError as exc:
                    print(exc)

            conf = struct['configuration']
            self.conf = conf

            prcs = struct['process']
            self.prcs = prcs

            confKeys = list(conf.keys())
            prcsKeys = list(prcs.keys())


            parse top-level configuration key information
            if 'name' in confKeys:
                self.name = conf['name']
            else:
                raise AttributeError('provided yaml file does not have a name parameter in configuration')

            if 'region' in confKeys:
                shp = gpd.read_file(conf['region'])
                self.region = list(shp.bounds.values[0])

            elif 'country' in confKeys:
                world = gpd.read_file(gpd.datasets.get_path('naturalearth_lowres'))
                country = world[world.name == conf['country']]
                if len(country) >= 1:
                    self.region = list(country.bounds.values[0])
                else:
                    raise ValueError('could not parse selected country from world shapefile')

            elif 'boundingbox' in confKeys:
                # from shapely import geometry
                # self.region = gpd.GeoDataFrame(pd.DataFrame({'id':[0],'geometry':[geometry.box(*conf['boundingbox'])]}))
                self.region = conf['boundingbox']

            else:
                raise AttributeError('provided yaml file does not have a specified region in configuration')

            if 'credentials' in confKeys:
                self.credentials = conf['credentials']
                self.earthdataLogin = self.credentials['earthdata'].values()
            else:
                self.credentials = None

            if 'stagingBucket' in confKeys:
                self.stagingBucket = conf['stagingBucket']
            else:
                self.stagingBucket = None

            if 'targetAsset' in confKeys:
                self.targetAsset = conf['targetAsset']
                # if ~self.targetAsset.endswith('/'):
                #     self.targetAsset += '/'
            else:
                self.targetAsset = None

            if 'workdir' in confKeys:
                self.workdir = conf['workdir']
            else:
                warning.warn("working dir for processing could not be parsed from provided yaml file, setting working dir to current dir",
                            UserWarning)
                self.workdir = './'

            # parse processing specific key information
            if 'hand' in prcsKeys:
                self.hand = prcs['hand']
            else:
                warning.warn("HAND assetID could not be parsed from provided yaml file, setting default global hand model",
                            UserWarning)
                self.hand = 'users/gena/GlobalHAND/30m/hand-5000'

            if 'atms' in prcsKeys:
                self.atmsParams = prcs['atms']

            if 'viirs' in prcsKeys:
                self.viirsParams = prcs['viirs']

            if 'sentinel1' in prcsKeys:
                self.s1Params = prcs['sentinel1']

        else:
            raise ValueError('yamlfile configuration argument must be specified')

        return


    def process(self,product, date,skipPreprocessing=False):
        try:
            ee.Initialize()
        except EEException as e:
            print(e)

        self._parse_config()

        if product in ['sentinel1','atms','viirs','modis']:
            dt = utils.decode_date(date)
            tomorrow = (dt + datetime.timedelta(1)).strftime('%Y-%m-%d')

            dateDir = os.path.join(self.workdir,dt.strftime('%Y%m%d'))
            prodDir = os.path.join(dateDir,product)

            geom = ee.Geometry.Rectangle(self.region)

            hand = ee.Image(self.hand)

            if product == 'atms':
                params = self.atmsParams
                paramKeys = list(params.keys())

                collId = self.atmsParams['waterFractionAsset']
                worker = Atms(geom,date,tomorrow,collectionid=collId)

                if skipPreprocessing == False:
                    if os.path.exists(dateDir) != True:
                        os.mkdir(dateDir)
                    if os.path.exists(prodDir) != True:
                        os.mkdir(prodDir)

                    geotiffs = worker.extract(dt,self.region,credentials=self.earthdataLogin,outDir=prodDir,gridding_radius=50000)
                    worker.load(geotiffs,self.stagingBucket,collId)

                if 'seed' in paramKeys:
                    permanentWater = ee.Image(self.atmsParams['seed'])
                else:
                    permanentWater = None

                if 'probablistic' in paramKeys:
                    runProbs = params['probablistic']
                else:
                    runProbs = False

                waterImage = worker.waterMap(hand,permanent=permanentWater,probablistic=runProbs)
                waterImage = waterImage.set({'system:time_start':ee.Date(date).millis(),'sensor':product})
                assetTarget = self.targetAsset + '{0}_bathtub_{1}'.format(product,date.replace('-',''))

            elif (product == 'viirs') or (product == 'modis'):
                today = datetime.datetime.now()

                if (today - dt).days < 5:
                    avail = today - datetime.timedelta(5)
                    raise NotImplementedError('NRT processing for VIIRS or MODIS has not been implemented, please select a date prior to {}'.format(avail))
                else:
                    minDate = (dt - datetime.timedelta(45)).strftime('%Y-%m-%d')
                    maxDate = (dt + datetime.timedelta(1)).strftime('%Y-%m-%d')

                    if product == 'modis':
                        worker = Modis(geom,minDate,maxDate,collectionid='MODIS/006/MOD09GA')
                        params = self.viirsParams
                    else:
                        worker = Viirs(geom,minDate,maxDate,collectionid='NOAA/VIIRS/001/VNP09GA')
                        params = self.viirsParams

                    paramKeys = list(params.keys())

                    ls = Landsat(geom,minDate,maxDate,collectionid='LANDSAT/LC08/C01/T1_SR')
                    s2 = Sentinel2(geom,minDate,maxDate,collectionid='COPERNICUS/S2_SR')
                    highRes = ee.ImageCollection(ls.collection.merge(s2.collection))

                    worker.downscale(highRes,target_date=date,windowSize=33,A=0.5)

                    if 'probablistic' in paramKeys:
                        runProbs = params['probablistic']
                        params.pop('probablistic')
                    else:
                        runProbs = False
                        nIters=100
                        probTreshold=0.75

                    waterImage = worker.waterMap(date,hand,probablistic=runProbs,**params)
                    waterImage = waterImage\
                        .set({'system:time_start':ee.Date(date).millis(),'sensor':product})
                    assetTarget = self.targetAsset + '{0}_downscaled_globalOtsu_{1}'.format(product,date.replace('-',''))

            elif product == 'sentinel1':
                previous = (dt + datetime.timedelta(0)).strftime('%Y-%m-%d')
                worker = Sentinel1(geom,previous,tomorrow,collectionid='COPERNICUS/S1_GRD')
                waterImage = worker.waterMap(date).And(hand.lt(30))
                waterImage = waterImage.rename('water')\
                    .set({'system:time_start':ee.Date(date).millis(),'sensor':product})
                assetTarget = self.targetAsset + '{0}_bootstrapOtsu_{1}'.format(product,date.replace('-',''))

            else:
                raise NotImplementedError('select product is currently not implemented, please check back with later versions')

            description = '{0}_water_{1}'.format(product,date)
            geeutils.exportImage(waterImage,geom,assetTarget,description=description)

        else:
            raise NotImplementedError('select product is currently not implemented, please check back with later versions')

        return

    def run_tests(self):
        raise NotImplementedError('test functionality not implemented...please ')
        return

    @staticmethod
    def init_env():
        # authenticate earth engine
        cmd = "earthengine authenticate"
        os.system(cmd)

        # initialize gcloud environment
        cmd = "gcloud init"
        os.system(cmd)

        return


def main():
    fire.Fire(hydrafloods)
    return



import ee
import copy
import numpy as np
import pandas as pd
from hydrafloods import geeutils, decorators


def fcnn(img, out_band_bames=None, probabilities=True, **kwargs):
    @decorators.carry_metadata
    def _apply(image):
        # run the predictions
        prediction = model.predictImage(image.toFloat().toArray())

        if probabilities is False:
            # find highest probability class
            precition = prediction.arrayArgmax().arrayFlatten([outBandNames])
        else:
            # flatten probability array to image with bands
            prediction = prediction.arrayFlatten([outBandNames]).toFloat()

        return prediction

    # Load the trained model and use it for prediction.
    model = ee.Model.fromAiPlatformPredictor(**kwargs)

    # run predictions over image collection and
    return _apply(img)


def logistic_regression():

    return


def fc_to_scaler(fc):
    return

def random_forest_ee(
    n_trees,
    feature_collection,
    feature_names,
    label,
    normalize_features=True,
    mode="classification",
):
    def feature_scaling(feature):
        def iter_cols(i):
            i = ee.String(i)
            v = ee.Number(feature.get(i))
            minv = ee.Number(min_max_dict.get(i.cat("_min")))
            maxv = ee.Number(min_max_dict.get(i.cat("_max")))
            return v.subtract(minv).divide(maxv.subtract(minv))

        scaled = ee_feature_names.map(iter_cols)
        newVals = ee.Dictionary.fromLists(ee_feature_names, scaled)
        return feature.set(newVals)

    fc = ee.FeatureCollection(feature_collection)
    ee_feature_names = ee.List(feature_names)

    if normalize_features:
        feature_min_max = fc.reduceColumns(
            ee.Reducer.minMax().repeat(ee_feature_names.length()), ee_feature_names
        )

        # min/max feature names
        names = ee_feature_names.map(
            lambda x: ee.List([ee.String(x).cat("_min"), ee.String(x).cat("_max")])
        ).flatten()

        # get the min/max values for each feature
        # used to scale values from 0-1
        min_max_dict = ee.Dictionary.fromLists(
            names,
            ee.List(feature_min_max.get("min"))
            .zip(feature_min_max.get("max"))
            .flatten(),
        )

        # normalize the features in the entire featureCollection
        fc_norm = fc.map(feature_scaling)

    else:
        fc_norm = fc

    classifier = (
        ee.Classifier.smileRandomForest(n_trees)
        .setOutputMode(mode.upper())
        .train(fc_norm, label, ee_feature_names)
    )

    return classifier, min_max_dict


def random_forest_fc(asset):
    if sample_path.startswith("gs://"):
        estimator_files = ee.List(utils.list_gcs_objs(path, pattern="estimator*.txt"))

        forest_strings = estimator_files.map(lambda x: ee.Blob(x).string())

        classifier = ee.Classifier.decisionTreeEnsemble(forest_strings)

    else:
        raise NotImplementedError("currently only reading files from GCS is supported")

    return classifier


def optimize(model, features, labels, opt_params=None):
    from bayes_opt import BayesianOptimization

    def crossval(n_estimators, min_samples_split, max_features):
        """Wrapper of RandomForest cross validation.
        Notice how we ensure n_estimators and min_samples_split are casted
        to integer before we pass them along. Moreover, to avoid max_features
        taking values outside the (0, 1) range, we also ensure it is capped
        accordingly.
        """
        return cv(
            n_estimators=int(n_estimators),
            min_samples_split=int(min_samples_split),
            max_features=max(min(max_features, 0.999), 1e-3),
            data=data,
            targets=targets,
        )

    optimizer = BayesianOptimization(
        f=crossval,
        pbounds={
            "n_estimators": (10, 250),
            "min_samples_split": (2, 25),
            "max_features": (0.1, 0.999),
        },
        random_state=1234,
        verbose=2,
    )
    optimizer.maximize(n_iter=10)

    print("Final result:", optimizer.max)

    return


def cv(model, features, labels, n_cv=5, classifier=False, **kwargs):
    """Random Forest cross validation.
    This function will instantiate a random forest classifier with parameters
    n_estimators, min_samples_split, and max_features. Combined with data and
    targets this will in turn be used to perform cross validation. The result
    of cross validation is returned.
    Our goal is to find combinations of n_estimators, min_samples_split, and
    max_features that minimzes the log loss.
    """

    score_metric = "neg_log_loss" if classifier else "mean_squared_error"
    estimator = model(**kwargs)
    cval = cross_val_score(estimator, features, labels, scoring=score_metric, cv=n_cv)
    return cval.mean()


def sklearn_tree_to_string(estimator, feature_names):

    # extract out the information need to build the tree string
    n_nodes = estimator.tree_.node_count
    children_left = estimator.tree_.children_left
    children_right = estimator.tree_.children_right
    feature_idx = estimator.tree_.feature
    values = estimator.tree_.value
    impurities = estimator.tree_.impurity
    n_samples = estimator.tree_.n_node_samples
    thresholds = estimator.tree_.threshold
    features = [feature_names[i] for i in feature_idx]

    # use iterative pre-order search to extract node depth and leaf information
    node_ids = np.zeros(shape=n_nodes, dtype=np.int64)
    node_depth = np.zeros(shape=n_nodes, dtype=np.int64)
    is_leaves = np.zeros(shape=n_nodes, dtype=bool)
    stack = [(0, -1)]  # seed is the root node id and its parent depth
    while len(stack) > 0:
        node_id, parent_depth = stack.pop()
        node_depth[node_id] = parent_depth + 1
        node_ids[node_id] = node_id

        # If we have a test node
        if children_left[node_id] != children_right[node_id]:
            stack.append((children_left[node_id], parent_depth + 1))
            stack.append((children_right[node_id], parent_depth + 1))
        else:
            is_leaves[node_id] = True

    # create a table of the initial structure
    # each row is a node or leaf
    df = pd.DataFrame(
        {
            "node_id": node_ids,
            "node_depth": node_depth,
            "is_leaf": is_leaves,
            "children_left": children_left,
            "children_right": children_right,
            "value": np.squeeze(values),
            "criterion": impurities,
            "n_samples": n_samples,
            "threshold": thresholds,
            "feature_name": features,
            "sign": ["<="] * n_nodes,
        }
    )

    # the table representation does not have lef vs right node structure
    # so we need to add in right nodes in the correct location
    # we do this by first calculating which nodes are right and then insert them at the correct index

    # get a dict of right node rows and assign key based on index where to insert
    inserts = {}
    for row in df.itertuples():
        child_r = row.children_right
        if child_r > row.Index:
            ordered_row = np.array(row)
            ordered_row[-1] = ">"
            inserts[child_r] = ordered_row[1:]  # drop index value
    # sort the inserts as to keep track of the additive indexing
    inserts_sorted = {k: inserts[k] for k in sorted(inserts.keys())}

    # loop through the row inserts and add to table (array)
    table_values = df.values
    for i, k in enumerate(inserts_sorted.keys()):
        table_values = np.insert(table_values, (k + i), inserts_sorted[k], axis=0)

    # make the ordered table array into a dataframe
    # note: df is dtype "object", need to cast later on
    ordered_df = pd.DataFrame(table_values, columns=df.columns)

    max_depth = np.max(ordered_df.node_depth.astype(int))
    tree_str = f"1) root 0 0 0 (0)\n"
    previous_depth = -1
    cnts = []
    # loop through the nodes and calculate the node number and values per node
    for row in ordered_df.itertuples():
        node_depth = int(row.node_depth)
        left = int(row.children_left)
        right = int(row.children_right)
        if left != right:
            if row.Index == 0:
                cnt = 2
            elif previous_depth > node_depth:
                depths = ordered_df.node_depth.values[: row.Index]
                idx = np.where(depths == node_depth)[0][-1]
                # cnt = (cnts[row.Index-1] // 2) + 1
                cnt = cnts[idx] + 1
            elif previous_depth < node_depth:
                cnt = cnts[row.Index - 1] * 2
            elif previous_depth == node_depth:
                cnt = cnts[row.Index - 1] + 1

            if node_depth == (max_depth - 1):
                value = float(ordered_df.iloc[row.Index + 1].value)
                samps = int(ordered_df.iloc[row.Index + 1].n_samples)
                criterion = float(ordered_df.iloc[row.Index + 1].criterion)
                tail = " *\n"
            else:
                if (
                    (bool(ordered_df.loc[ordered_df.node_id == left].iloc[0].is_leaf))
                    and (
                        bool(
                            int(row.Index)
                            < int(ordered_df.loc[ordered_df.node_id == left].index[0])
                        )
                    )
                    and (str(row.sign) == "<=")
                ):
                    rowx = ordered_df.loc[ordered_df.node_id == left].iloc[0]
                    tail = " *\n"
                    value = float(rowx.value)
                    samps = int(rowx.n_samples)
                    criterion = float(rowx.criterion)

                elif (
                    (bool(ordered_df.loc[ordered_df.node_id == right].iloc[0].is_leaf))
                    and (
                        bool(
                            int(row.Index)
                            < int(ordered_df.loc[ordered_df.node_id == right].index[0])
                        )
                    )
                    and (str(row.sign) == ">")
                ):
                    rowx = ordered_df.loc[ordered_df.node_id == right].iloc[0]
                    tail = " *\n"
                    value = float(rowx.value)
                    samps = int(rowx.n_samples)
                    criterion = float(rowx.criterion)

                else:
                    value = float(row.value)
                    samps = 0  # int(row.n_samples)
                    criterion = 0  # float(row.criterion)
                    tail = "\n"

            # extract out the information needed in each line
            # spacing = (node_depth + 1) * "  "  # for pretty printing
            fname = str(row.feature_name)  # name of the feature (i.e. band name)
            tresh = float(row.threshold)  # threshold
            sign = str(row.sign)

            # tree_str += f"{cnt}) {fname} {sign} {tresh:.6f} {samps} {criterion:.4f} {value:.6f}{tail}"
            tree_str += f"{cnt}) {fname} {sign} {tresh:.4f} 0 0 {value:.4f}{tail}"
            previous_depth = node_depth
        cnts.append(cnt)

    return tree_str




import ee
from ee.ee_exception import EEException
import math
import string
import random
import datetime
from hydrafloods import decorators


# helper function to convert qa bit image to flag
def extract_bits(image, start, end=None, new_name=None):
    newame = new_name if new_name is not None else f'{start}Bits'

    if (start == end) or (end is None):
        #perform a bit shift with bitwiseAnd
        return image.select([0],[new_name])\
            .bitwiseAnd(1<<start)
    else:
        # Compute the bits we need to extract.
        pattern = 0
        for i in range(start, end):
            pattern += int(math.pow(2, i))

        # Return a single band image of the extracted QA bits, giving the band
        # a new name.
        return image.select([0], [new_name])\
            .bitwiseAnd(pattern)\
            .rightShift(start)


def get_geoms(img):
    return img.geometry()


def get_tile_layer_url(ee_image_object):
    map_id = ee.Image(ee_image_object).getMapId()
    tile_url_template = (
        "https://earthengine.googleapis.com/map/{mapid}/{{z}}/{{x}}/{{y}}?token={token}"
    )
    return tile_url_template.format(**map_id)


def export_image(
    image,
    region,
    asset_id,
    description=None,
    scale=1000,
    crs="EPSG:4326",
    pyramiding=None,
):
    if (description == None) or (type(description) != str):
        description = "".join(
            random.SystemRandom().choice(string.ascii_letters) for _ in range(8)
        ).lower()
    # get serializable geometry for export
    export_region = region.bounds(maxError=1).getInfo()["coordinates"]

    if pyramiding is None:
        pyramiding = {".default": "mean"}

    # set export process
    export = ee.batch.Export.image.toAsset(
        image,
        description=description,
        assetId=asset_id,
        scale=scale,
        region=export_region,
        maxPixels=1e13,
        crs=crs,
        pyramidingPolicy=pyramiding,
    )
    # start export process
    export.start()

    return


def export_table(
    table,
    region,
    assetId,
    description=None,
    scale=1000,
    crs="EPSG:4326",
    pyramiding=None,
):
    if (description == None) or (type(description) != str):
        description = "".join(
            random.SystemRandom().choice(string.ascii_letters) for _ in range(8)
        ).lower()
    # get serializable geometry for export
    exportRegion = region.bounds(maxError=1).getInfo()["coordinates"]

    if pyramiding is None:
        pyramiding = {".default": "mean"}

    # set export process
    export = ee.batch.Export.image.toAsset(
        image,
        description=description,
        assetId=assetId,
        scale=scale,
        region=exportRegion,
        maxPixels=1e13,
        crs=crs,
        pyramidingPolicy=pyramiding,
    )
    # start export process
    export.start()

    return


def batch_export(
    collection,
    collection_asset,
    bucket=None,
    region=None,
    prefix=None,
    suffix=None,
    scale=1000,
    crs="EPSG:4326",
    metadata=None,
    pyramiding=None,
    verbose=False,
):
    if type(collection) is not ee.imagecollection.ImageCollection:
        try:
            collection = getattr(collection, "collection")
        except Exception as e:
            raise TypeError(
                "argument collection needs to be either of type ee.ImageCollection "
                "or hydrafloods.hfCollection"
            )

    n = collection.size()
    exportImages = collection.sort("system:time_start", False).toList(n)
    nIter = n.getInfo()

    for i in range(nIter):
        img = ee.Image(exportImages.get(i))
        if metadata is not None:
            img = img.set(metadata)

        t = img.get("system:time_start").getInfo()
        date = datetime.datetime.utcfromtimestamp(t / 1e3).strftime("%Y%m%d")

        if region is None:
            region = img.geometry()

        exportName = date
        if prefix is not None:
            exportName = f"{prefix}_" + exportName
        if suffix is not None:
            exportName = exportName + f"_{suffix}"

        description = exportName
        if verbose:
            print(f"running export for {description}")

        if not collection_asset.endswith("/"):
            collection_asset += "/"

        exportName = collection_asset + description

        export_image(
            img,
            region,
            exportName,
            description=description,
            scale=scale,
            crs=crs,
            pyramiding=pyramiding,
        )

    return


@decorators.carry_metadata
def rescale_bands(img):
    def _individual_band(b):
        b = ee.String(b)
        minKey = b.cat("_min")
        maxKey = b.cat("_max")
        return img.select(b).unitScale(
            ee.Number(minMax.get(minKey)), ee.Number(minMax.get(maxKey))
        )

    bandNames = img.bandNames()
    geom = img.geometry()

    minMax = img.reduceRegion(
        reducer=ee.Reducer.minMax(), geometry=geom, scale=90, bestEffort=True
    )

    rescaled = bandNames.map(_individual_band)
    bandSequence = ee.List.sequence(0, bandNames.length().subtract(1))

    return ee.ImageCollection(rescaled).toBands().select(bandSequence, bandNames)


@decorators.carry_metadata
def power_to_db(img):
    return ee.Image(10).multiply(img.log10())


@decorators.carry_metadata
def db_to_power(img):
    return ee.Image(10).pow(img.divide(10))


@decorators.carry_metadata
def add_indices(img):
    ndvi = img.normalizedDifference(["nir", "red"]).rename("ndvi")
    mndwi = img.normalizedDifference(["green", "swir1"]).rename("mndwi")
    # nwi = img.expression(
    #     "((b-(n+s+w))/(b+(n+s+w))*100)",
    #     {
    #         "b": img.select("blue"),
    #         "n": img.select("nir"),
    #         "s": img.select("swir1"),
    #         "w": img.select("swir2"),
    #     },
    # ).rename("nwi")
    # aewinsh = img.expression(
    #     "4.0 * (g-s) - ((0.25*n) + (2.75*w))",
    #     {
    #         "g": img.select("green"),
    #         "s": img.select("swir1"),
    #         "n": img.select("nir"),
    #         "w": img.select("swir2"),
    #     },
    # ).rename("aewinsh")
    # aewish = img.expression(
    #     "b+2.5*g-1.5*(n+s)-0.25*w",
    #     {
    #         "b": img.select("blue"),
    #         "g": img.select("green"),
    #         "n": img.select("nir"),
    #         "s": img.select("swir1"),
    #         "w": img.select("swir2"),
    #     },
    # ).rename("aewish")
    # tcwet = img.expression(
    #     "0.1509*b + 0.1973*g + 0.3279*r + 0.3406*n - 0.7112*s - 0.4572*w",
    #     {
    #         "b": img.select("blue"),
    #         "g": img.select("green"),
    #         "r": img.select("red"),
    #         "n": img.select("nir"),
    #         "s": img.select("swir1"),
    #         "w": img.select("swir2"),
    #     },
    # ).rename("tcwet")

    out_img = ee.Image.cat([img,ndvi, mndwi])#, nwi, aewinsh, aewish, tcwet])

    return out_img

def tile_region(region,grid_size=0.1,intersect_geom=None,contain_geom=None):
    def constuctGrid(i):
        def contructXGrid(j):
            j = ee.Number(j)
            box = ee.Feature(
                ee.Geometry.Rectangle([j, i, j.add(grid_size), i.add(grid_size)],"epsg:4326",geodesic=False)
            )
            if contain_geom is not None:
                out = ee.Algorithms.If(
                    region.contains(contain_geom, maxError=10), box, None
                )
            elif intersect_geom is not None:
                out = ee.Algorithms.If(
                    region.intersects(intersect_geom, maxError=10), box, None
                )
            else:
                out = box
            return ee.Feature(out)

        i = ee.Number(i)
        out = ee.List.sequence(west, east.subtract(grid_size), grid_size).map(
            contructXGrid
        )
        return out

    if (contain_geom is not None) and (intersect_geom is not None):
        raise ValueError("contains and intersection keywords are mutually exclusive, please define only one")

    bounds = region.bounds(maxError=100)
    coords = ee.List(bounds.coordinates().get(0))
    grid_res = ee.Number(grid_size)

    west = ee.Number(ee.List(coords.get(0)).get(0))
    south = ee.Number(ee.List(coords.get(0)).get(1))
    east = ee.Number(ee.List(coords.get(2)).get(0))
    north = ee.Number(ee.List(coords.get(2)).get(1))

    west = west.subtract(west.mod(grid_res))
    south = south.subtract(south.mod(grid_res))
    east = east.add(grid_res.subtract(east.mod(grid_res)))
    north = north.add(grid_res.subtract(north.mod(grid_res)))

    grid = ee.FeatureCollection(
        ee.List.sequence(south, north.subtract(grid_res), grid_res)
        .map(constuctGrid)
        .flatten()
    )

    return grid


import ee
import math
from hydrafloods import geeutils, decorators


@decorators.carry_metadata
def lee_sigma(img, window=9, sigma=0.9, looks=4, tk=7,keep_bands="angle"):
    band_names = img.bandNames()
    proc_bands = band_names.remove(keep_bands)
    keep_img = img.select(keep_bands)
    img = img.select(proc_bands)

    midPt = (window // 2) + 1 if (window % 2) != 0 else window // 2
    kernelWeights = ee.List.repeat(ee.List.repeat(1, window), window)
    kernel = ee.Kernel.fixed(window, window, kernelWeights, midPt, midPt)

    targetWeights = ee.List.repeat(ee.List.repeat(1, 3), 3)
    targetkernel = ee.Kernel.fixed(3, 3, targetWeights, 1, 1)

    # Lookup table for range and eta values for intensity
    sigmaLookup = ee.Dictionary(
        {
            1: ee.Dictionary(
                {
                    0.5: ee.Dictionary({"A1": 0.436, "A2": 1.92, "η": 0.4057}),
                    0.6: ee.Dictionary({"A1": 0.343, "A2": 2.21, "η": 0.4954}),
                    0.7: ee.Dictionary({"A1": 0.254, "A2": 2.582, "η": 0.5911}),
                    0.8: ee.Dictionary({"A1": 0.168, "A2": 3.094, "η": 0.6966}),
                    0.9: ee.Dictionary({"A1": 0.084, "A2": 3.941, "η": 0.8191}),
                    0.95: ee.Dictionary({"A1": 0.043, "A2": 4.840, "η": 0.8599}),
                }
            ),
            2: ee.Dictionary(
                {
                    0.5: ee.Dictionary({"A1": 0.582, "A2": 1.584, "η": 0.2763}),
                    0.6: ee.Dictionary({"A1": 0.501, "A2": 1.755, "η": 0.3388}),
                    0.7: ee.Dictionary({"A1": 0.418, "A2": 1.972, "η": 0.4062}),
                    0.8: ee.Dictionary({"A1": 0.327, "A2": 2.260, "η": 0.4819}),
                    0.9: ee.Dictionary({"A1": 0.221, "A2": 2.744, "η": 0.5699}),
                    0.95: ee.Dictionary({"A1": 0.152, "A2": 3.206, "η": 0.6254}),
                }
            ),
            3: ee.Dictionary(
                {
                    0.5: ee.Dictionary({"A1": 0.652, "A2": 1.458, "η": 0.2222}),
                    0.6: ee.Dictionary({"A1": 0.580, "A2": 1.586, "η": 0.2736}),
                    0.7: ee.Dictionary({"A1": 0.505, "A2": 1.751, "η": 0.3280}),
                    0.8: ee.Dictionary({"A1": 0.419, "A2": 1.865, "η": 0.3892}),
                    0.9: ee.Dictionary({"A1": 0.313, "A2": 2.320, "η": 0.4624}),
                    0.95: ee.Dictionary({"A1": 0.238, "A2": 2.656, "η": 0.5084}),
                }
            ),
            4: ee.Dictionary(
                {
                    0.5: ee.Dictionary({"A1": 0.694, "A2": 1.385, "η": 0.1921}),
                    0.6: ee.Dictionary({"A1": 0.630, "A2": 1.495, "η": 0.2348}),
                    0.7: ee.Dictionary({"A1": 0.560, "A2": 1.627, "η": 0.2825}),
                    0.8: ee.Dictionary({"A1": 0.480, "A2": 1.804, "η": 0.3354}),
                    0.9: ee.Dictionary({"A1": 0.378, "A2": 2.094, "η": 0.3991}),
                    0.95: ee.Dictionary({"A1": 0.302, "A2": 2.360, "η": 0.4391}),
                }
            ),
        }
    )

    # extract data from lookup
    looksDict = ee.Dictionary(sigmaLookup.get(ee.String(str(looks))))
    sigmaImage = ee.Dictionary(looksDict.get(ee.String(str(sigma)))).toImage()
    a1 = sigmaImage.select("A1")
    a2 = sigmaImage.select("A2")
    aRange = a2.subtract(a1)
    eta = sigmaImage.select("η").pow(2)

    img = geeutils.db_to_power(img)

    # MMSE estimator
    mmseMask = img.gte(a1).Or(img.lte(a2))
    mmseIn = img.updateMask(mmseMask)
    oneImg = ee.Image(1)
    z = mmseIn.reduceNeighborhood(ee.Reducer.mean(), kernel, None, True)
    varz = mmseIn.reduceNeighborhood(ee.Reducer.variance(), kernel)
    varx = (varz.subtract(z.abs().pow(2).multiply(eta))).divide(oneImg.add(eta))
    b = varx.divide(varz)
    mmse = oneImg.subtract(b).multiply(z.abs()).add(b.multiply(mmseIn))

    # workflow
    z99 = ee.Dictionary(
        img.reduceRegion(
            reducer=ee.Reducer.percentile([99], None, 255, 0.001, 1e6),
            geometry=img.geometry(),
            scale=10,
            bestEffort=True,
        )
    ).toImage()

    overThresh = img.gte(z99)

    K = overThresh.reduceNeighborhood(ee.Reducer.sum(), targetkernel, None, True)

    retainPixel = K.gte(tk)
    xHat = geeutils.power_to_db(img.updateMask(retainPixel).unmask(mmse))

    return ee.Image(xHat).rename(proc_bands).addBands(keep_img)


# The RL speckle filter
@decorators.carry_metadata
def refined_lee(image):
    def apply_filter(b):
        img = power.select([b])

        # img must be in natural units, i.e. not in dB!
        # Set up 3x3 kernels
        weights3 = ee.List.repeat(ee.List.repeat(1, 3), 3)
        kernel3 = ee.Kernel.fixed(3, 3, weights3, 1, 1, False)

        mean3 = img.reduceNeighborhood(ee.Reducer.mean(), kernel3)
        variance3 = img.reduceNeighborhood(ee.Reducer.variance(), kernel3)

        # Use a sample of the 3x3 windows inside a 7x7 windows to determine gradients and directions
        sample_weights = ee.List(
            [
                [0, 0, 0, 0, 0, 0, 0],
                [0, 1, 0, 1, 0, 1, 0],
                [0, 0, 0, 0, 0, 0, 0],
                [0, 1, 0, 1, 0, 1, 0],
                [0, 0, 0, 0, 0, 0, 0],
                [0, 1, 0, 1, 0, 1, 0],
                [0, 0, 0, 0, 0, 0, 0],
            ]
        )

        sample_kernel = ee.Kernel.fixed(7, 7, sample_weights, 3, 3, False)

        # Calculate mean and variance for the sampled windows and store as 9 bands
        sample_mean = mean3.neighborhoodToBands(sample_kernel)
        sample_var = variance3.neighborhoodToBands(sample_kernel)

        # Determine the 4 gradients for the sampled windows
        gradients = sample_mean.select(1).subtract(sample_mean.select(7)).abs()
        gradients = gradients.addBands(
            sample_mean.select(6).subtract(sample_mean.select(2)).abs()
        )
        gradients = gradients.addBands(
            sample_mean.select(3).subtract(sample_mean.select(5)).abs()
        )
        gradients = gradients.addBands(
            sample_mean.select(0).subtract(sample_mean.select(8)).abs()
        )

        # And find the maximum gradient amongst gradient bands
        max_gradient = gradients.reduce(ee.Reducer.max())

        # Create a mask for band pixels that are the maximum gradient
        gradmask = gradients.eq(max_gradient)

        # duplicate gradmask bands: each gradient represents 2 directions
        gradmask = gradmask.addBands(gradmask)

        # Determine the 8 directions
        directions = (
            sample_mean.select(1)
            .subtract(sample_mean.select(4))
            .gt(sample_mean.select(4).subtract(sample_mean.select(7)))
            .multiply(1)
        )
        directions = directions.addBands(
            sample_mean.select(6)
            .subtract(sample_mean.select(4))
            .gt(sample_mean.select(4).subtract(sample_mean.select(2)))
            .multiply(2)
        )
        directions = directions.addBands(
            sample_mean.select(3)
            .subtract(sample_mean.select(4))
            .gt(sample_mean.select(4).subtract(sample_mean.select(5)))
            .multiply(3)
        )
        directions = directions.addBands(
            sample_mean.select(0)
            .subtract(sample_mean.select(4))
            .gt(sample_mean.select(4).subtract(sample_mean.select(8)))
            .multiply(4)
        )
        # The next 4 are the not() of the previous 4
        directions = directions.addBands(directions.select(0).Not().multiply(5))
        directions = directions.addBands(directions.select(1).Not().multiply(6))
        directions = directions.addBands(directions.select(2).Not().multiply(7))
        directions = directions.addBands(directions.select(3).Not().multiply(8))

        # Mask all values that are not 1-8
        directions = directions.updateMask(gradmask)

        # "collapse" the stack into a singe band image (due to masking, each pixel has just one value (1-8) in it's directional band, and is otherwise masked)
        directions = directions.reduce(ee.Reducer.sum())

        sample_stats = sample_var.divide(sample_mean.multiply(sample_mean))

        # Calculate localNoiseVariance
        sigmaV = (
            sample_stats.toArray()
            .arraySort()
            .arraySlice(0, 0, 5)
            .arrayReduce(ee.Reducer.mean(), [0])
        )

        # Set up the 7*7 kernels for directional statistics
        rect_weights = ee.List.repeat(ee.List.repeat(0, 7), 3).cat(
            ee.List.repeat(ee.List.repeat(1, 7), 4)
        )

        diag_weights = ee.List(
            [
                [1, 0, 0, 0, 0, 0, 0],
                [1, 1, 0, 0, 0, 0, 0],
                [1, 1, 1, 0, 0, 0, 0],
                [1, 1, 1, 1, 0, 0, 0],
                [1, 1, 1, 1, 1, 0, 0],
                [1, 1, 1, 1, 1, 1, 0],
                [1, 1, 1, 1, 1, 1, 1],
            ]
        )

        rect_kernel = ee.Kernel.fixed(7, 7, rect_weights, 3, 3, False)
        diag_kernel = ee.Kernel.fixed(7, 7, diag_weights, 3, 3, False)

        # Create stacks for mean and variance using the original kernels. Mask with relevant direction.
        dir_mean = img.reduceNeighborhood(ee.Reducer.mean(), rect_kernel).updateMask(
            directions.eq(1)
        )
        dir_var = img.reduceNeighborhood(ee.Reducer.variance(), rect_kernel).updateMask(
            directions.eq(1)
        )

        dir_mean = dir_mean.addBands(
            img.reduceNeighborhood(ee.Reducer.mean(), diag_kernel).updateMask(
                directions.eq(2)
            )
        )
        dir_var = dir_var.addBands(
            img.reduceNeighborhood(ee.Reducer.variance(), diag_kernel).updateMask(
                directions.eq(2)
            )
        )

        # and add the bands for rotated kernels
        for i in range(1, 4):
            dir_mean = dir_mean.addBands(
                img.reduceNeighborhood(
                    ee.Reducer.mean(), rect_kernel.rotate(i)
                ).updateMask(directions.eq(2 * i + 1))
            )
            dir_var = dir_var.addBands(
                img.reduceNeighborhood(
                    ee.Reducer.variance(), rect_kernel.rotate(i)
                ).updateMask(directions.eq(2 * i + 1))
            )
            dir_mean = dir_mean.addBands(
                img.reduceNeighborhood(
                    ee.Reducer.mean(), diag_kernel.rotate(i)
                ).updateMask(directions.eq(2 * i + 2))
            )
            dir_var = dir_var.addBands(
                img.reduceNeighborhood(
                    ee.Reducer.variance(), diag_kernel.rotate(i)
                ).updateMask(directions.eq(2 * i + 2))
            )

        # "collapse" the stack into a single band image (due to masking, each pixel has just one value in it's directional band, and is otherwise masked)
        dir_mean = dir_mean.reduce(ee.Reducer.sum())
        dir_var = dir_var.reduce(ee.Reducer.sum())

        # A finally generate the filtered value
        varX = dir_var.subtract(dir_mean.multiply(dir_mean).multiply(sigmaV)).divide(
            sigmaV.add(1.0)
        )

        b = varX.divide(dir_var)

        # return multi-band image band from array
        return (
            dir_mean.add(b.multiply(img.subtract(dir_mean)))
            .arrayProject([0])
            .arrayFlatten([["sum"]])
            .float()
        )

    bandNames = image.bandNames()
    power = geeutils.db_to_power(image)

    result = ee.ImageCollection(bandNames.map(apply_filter)).toBands().rename(bandNames)
    return geeutils.power_to_db(ee.Image(result))


@decorators.carry_metadata
def gamma_map(img, window=7, enl=5):

    bandNames = img.bandNames()
    # Square kernel, window should be odd (typically 3, 5 or 7)
    weights = ee.List.repeat(ee.List.repeat(1, window), window)
    midPt = (window // 2) + 1 if (window % 2) != 0 else window // 2

    # ~~(window/2) does integer division in JavaScript
    kernel = ee.Kernel.fixed(window, window, weights, midPt, midPt, False)

    # Convert image from dB to natural values
    nat_img = geeutils.db_to_power(img)

    # Get mean and variance
    mean = nat_img.reduceNeighborhood(ee.Reducer.mean(), kernel)
    variance = nat_img.reduceNeighborhood(ee.Reducer.variance(), kernel)

    # "Pure speckle" threshold
    ci = variance.sqrt().divide(mean)  # square root of inverse of enl

    # If ci <= cu, the kernel lies in a "pure speckle" area -> return simple mean
    cu = 1.0 / math.sqrt(enl)

    # If cu < ci < cmax the kernel lies in the low textured speckle area -> return the filtered value
    cmax = math.sqrt(2.0) * cu

    alpha = ee.Image(1.0 + cu * cu).divide(ci.multiply(ci).subtract(cu * cu))
    b = alpha.subtract(enl + 1.0)
    d = (
        mean.multiply(mean)
        .multiply(b)
        .multiply(b)
        .add(alpha.multiply(mean).multiply(nat_img).multiply(4.0 * enl))
    )
    f = b.multiply(mean).add(d.sqrt()).divide(alpha.multiply(2.0))

    caster = ee.Dictionary.fromLists(bandNames, ee.List.repeat("float", 3))
    img1 = (
        geeutils.power_to_db(mean.updateMask(ci.lte(cu))).rename(bandNames).cast(caster)
    )
    img2 = (
        geeutils.power_to_db(f.updateMask(ci.gt(cu)).updateMask(ci.lt(cmax)))
        .rename(bandNames)
        .cast(caster)
    )
    img3 = img.updateMask(ci.gte(cmax)).rename(bandNames).cast(caster)

    # If ci > cmax do not filter at all (i.e. we don't do anything, other then masking)
    result = (
        ee.ImageCollection([img1, img2, img3])
        .reduce(ee.Reducer.firstNonNull())
        .rename(bandNames)
        .clip(img.geometry())
    )

    # Compose a 3 band image with the mean filtered "pure speckle", the "low textured" filtered and the unfiltered portions
    return result



import os
import numpy as np
import xarray as xr
from scipy import ndimage, interpolate
from osgeo import gdal, osr
from pyproj import Proj, transform
from pyresample import bilinear, geometry, utils


def viirs(infile):
    tree = "//HDFEOS/GRIDS/VNP_Grid_{}_2D/Data_Fields/"
    field = "SurfReflect_{0}{1}_1"
    base = 'HDF5:"{0}":{1}{2}'

    outEpsg = 3857
    nd = -999

    # sinuWkt = '''PROJCS["Sphere_Sinusoidal",GEOGCS["GCS_Sphere",DATUM["D_Sphere",SPHEROID["Sphere",6371000.0,0.0]],PRIMEM["Greenwich",0.0],UNIT["Degree",0.0174532925199433]],PROJECTION["Sinusoidal"],PARAMETER["False_Easting",0.0],PARAMETER["False_Northing",0.0],PARAMETER["Central_Meridian",0.0],UNIT["Meter",1.0]]'''

    proj = "+proj=sinu +lon_0=0 +x_0=0 +y_0=0 +datum=WGS84 +ellps=WGS84 +a=6371007.181 +b=6371007.181 +units=m +no_defs"
    outProj = Proj(proj)
    inProj = Proj(init="epsg:4326")

    srs = osr.SpatialReference()
    srs.ImportFromProj4(outProj.definition_string())

    m = [i for i in range(12) if i not in [0, 6, 9]]
    i = [i for i in range(1, 4)]
    bands = [m, i]
    flatBands = [item for sublist in bands for item in sublist]

    res = ["1km", "500m"]
    mode = ["M", "I"]

    band = gdal.Open(base.format(infile, tree.format("1km"), field.format("QF", 1)))
    metadata = band.GetMetadata()
    # print(metadata)
    cloudQA = _extractBits(band.ReadAsArray(), 2, 3)
    hiresCloudQA = ndimage.zoom(cloudQA, 2, order=0)
    band = None

    band = gdal.Open(base.format(infile, tree.format("1km"), field.format("QF", 2)))
    shadowQA = _extractBits(band.ReadAsArray(), 3, 3)
    hiresShadowQA = ndimage.zoom(shadowQA, 2, order=0)

    qa = ~(hiresCloudQA > 0) & (hiresShadowQA < 1)

    ringLatitude = metadata["GRingLatitude"].split(" ")[:-1]
    ringLongitude = metadata["GRingLongitude"].split(" ")[:-1]

    ll, ul, ur, lr = [
        transform(inProj, outProj, float(ringLongitude[i]), float(ringLatitude[i]))
        for i in range(len(ringLatitude))
    ]
    iniX, iniY = ul[0], ul[1]
    print(ll, ul, ur, lr)

    bandNames = [
        "{0}{1}".format(mode[i], bands[i][j])
        for i in range(len(res))
        for j in range(len(bands[i]))
    ]

    subdata = [
        [infile, res[i], mode[i], bands[i][j]]
        for i in range(len(res))
        for j in range(len(bands[i]))
    ]

    data = np.zeros([qa.shape[0], qa.shape[1], len(subdata) + 1])

    for i, s in enumerate(subdata):
        infile, r, m, b = s

        subdataset = base.format(infile, tree.format(r), field.format(m, b))

        band = gdal.Open(subdataset)
        if m == "M":
            data = ndimage.zoom(band.ReadAsArray(), 2, order=0)

        else:
            sd = np.array(band.ReadAsArray())
        band = None

        sd[np.where(sd < 0)] = nd

        data[:, :, i] = sd

    data[:, :, -1] = qa

    res = float(metadata["CharacteristicBinSize500M"])
    yskew = math.sin(_getSkew(ur, ul)) * res
    xskew = math.cos(_getSkew(ul, ll)) * res
    print(xskew, yskew)
    yDim, xDim = qa.shape
    gt = (10007554.677, res, 0, 2223901.039333, 0, -res)
    print(gt)

    name, _ = os.path.splitext(infile)
    out_name = name + ".TIF"

    write_geotiff(out_name, data, gt, outEpsg, no_data=nd)

    return out_name


def atms(
    infile, gridding_radius=25000,
):
    ds = xr.open_dataset(infile)

    outEpsg = 3857
    nd = -999

    outProj = Proj(init="epsg:{0}".format(outEpsg))
    inProj = Proj(init="epsg:4326")

    lons, lats = ds.lon.values, ds.lat.values

    xx, yy = transform(inProj, outProj, lons, lats)
    minx, miny = transform(inProj, outProj, -180, -86)
    maxx, maxy = transform(inProj, outProj, 180, 86)
    res = 16000

    eastings = np.arange(round(minx), round(maxx), res)
    northings = np.arange(round(miny), round(maxy), res)

    ee = eastings[np.where((eastings > xx.min()) & (eastings < xx.max()))]
    nn = northings[np.where((northings > yy.min()) & (northings < yy.max()))]

    swath_def = geometry.SwathDefinition(lons=lons, lats=lats)
    area_def = geometry.AreaDefinition(
        "mercator",
        "WGS 84 / Pseudo-Mercator - Projected",
        "mercator",
        {
            "x_0": "0.0",
            "y_0": "0.0",
            "lat_ts": "0.00",
            "lon_0": "0.00",
            "proj": "merc",
            "k": "1.0",
            "datum": "WGS84",
            "ellps": "WGS84",
            "a": "6378137",
            "b": "6378137",
        },
        ee.size,
        nn.size,
        [ee.min(), nn.min(), ee.max(), nn.max()],
    )

    data = None

    # TODO: dynamically estimate sigama based on beam footprints
    eps = 0.1

    result = bilinear.resample_bilinear(
        ds.land_frac.where(ds["sat_zen"] < 50).values,
        swath_def,
        area_def,
        radius=gridding_radius,
        neighbours=32,
        nprocs=1,
        fill_value=nd,
        reduce_data=True,
        segments=None,
        epsilon=eps,
    )

    result[np.where(result >= 0)] = np.abs(result[np.where(result >= 0)] - 1) * 10000

    name, _ = os.path.splitext(infile)
    out_name = name + "_waterfrac.TIF"

    gt = (ee.min(), res, 0, nn.max(), 0, -res)

    write_geotiff(out_name, result, gt, outEpsg, no_data=nd)

    return out_name


def write_geotiff(out_name, data, gt, epsg, no_data=None):
    srs = osr.SpatialReference()
    srs.ImportFromEPSG(epsg)

    if len(data.shape) == 3:
        yDim, xDim = data.shape[:2]
        nBands = data.shape[2]

    elif len(data.shape) == 2:
        yDim, xDim = data.shape
        nBands = 1
    else:
        raise

    driver = gdal.GetDriverByName("GTiff")

    outDs = driver.Create(out_name, xDim, yDim, nBands, gdal.GDT_Int16)
    outDs.SetGeoTransform(gt)
    # set to something that is not user defined
    outDs.SetProjection(srs.ExportToWkt())

    for b in range(nBands):
        band = outDs.GetRasterBand(b + 1)
        if no_data:
            band.Setno_dataValue(-999)
        if nBands > 1:
            band.WriteArray(data[:, :, b])
        else:
            band.WriteArray(data)

        band = None

    outDs.FlushCache()

    return
